#include "vraymeshgeom.h"
#include "instance.h"
#include "pb2template_generator.h"
#include "defrayserver.h"
#include "defparams.h"
#include "vrender_unicode.h"

//************************************************************
// Class descriptor
//************************************************************

class VRayMeshGeomClassDesc: public ClassDesc2 {
public:
	int IsPublic(void) { return 1; }
	void *Create(BOOL loading) { return new VRayMeshGeom; }
	const TCHAR *ClassName(void) { return STR_CLASSNAME; }
	SClass_ID SuperClassID(void) { return GEOMOBJECT_CLASS_ID; }
	Class_ID ClassID(void) { return PLUGIN_CLASSID; }
	const TCHAR* Category(void) { return _T("VRay");  }

	// Hardwired name, used by MAX Script as unique identifier
	const TCHAR* InternalName(void) { return STR_INTERNALNAME; }
	HINSTANCE HInstance(void) { return hInstance; }
};

//************************************************************
// Static variables
//************************************************************

IObjParam* VRayMeshGeom::ip = NULL;
const float iconSize=40.0f;
TCHAR *iconText=_T("VRayMeshGeom");
static VRayMeshGeomClassDesc classDesc;

PluginManager *VRayMeshGeom::plugman;
Factory VRayMeshGeom::factory;
int VRayMeshGeom::numRenderPlugins=0;
int VRayMeshGeom::numFramePlugins=0;

static int shutdownCallbackRegistered=false;

//************************************************************
// DLL stuff
//************************************************************

HINSTANCE hInstance;
int controlsInit=FALSE;

BOOL WINAPI DllMain(HINSTANCE hinstDLL,ULONG fdwReason,LPVOID lpvReserved) {
	hInstance=hinstDLL;

	if (!controlsInit) {
		controlsInit=TRUE;
#if MAX_RELEASE<13900
		InitCustomControls(hInstance);
#endif
		InitCommonControls();
	}

	return(TRUE);
}

__declspec( dllexport ) const TCHAR* LibDescription() { return STR_LIBDESC; }
__declspec( dllexport ) int LibNumberClasses() { return 1; }

__declspec( dllexport ) ClassDesc* LibClassDesc(int i) {
	switch(i) { case 0: return &classDesc; }
	return NULL;
}

__declspec( dllexport ) ULONG LibVersion() { return VERSION_3DSMAX; }

//************************************************************
// Parameter block
//************************************************************

// Paramblock2 name
enum { params, }; 

static int ctrlID=100;

int nextID(void) { return ctrlID++; }

static ParamBlockDesc2 param_blk(params, STR_DLGTITLE,  0, &classDesc, P_AUTO_CONSTRUCT, REFNO_PBLOCK, 
	// Params
	pb_numSpheres, _T("num_spheres"), TYPE_INT, P_ANIMATABLE, 0,
		p_default, 10,
		p_range, 1, 10000,
		p_ui, TYPE_SPINNER, EDITTYPE_INT, nextID(), nextID(), SPIN_AUTOSCALE,
	PB_END,
	pb_radius, _T("radius"), TYPE_WORLD, P_ANIMATABLE, 0,
		p_default, 10.0f,
		p_range, 0.0f, 1e9f,
		p_ui, TYPE_SPINNER, EDITTYPE_UNIVERSE, nextID(), nextID(), SPIN_AUTOSCALE,
	PB_END,
	pb_numSegments, _T("num_segments"), TYPE_INT, P_ANIMATABLE, 0,
		p_default, 20,
		p_range, 3, 200,
		p_ui, TYPE_SPINNER, EDITTYPE_INT, nextID(), nextID(), SPIN_AUTOSCALE,
	PB_END,
	pb_boxSize, _T("box_size"), TYPE_WORLD, P_ANIMATABLE, 0,
		p_default, 100.0f,
		p_range, 1e-9f, 1e9f,
		p_ui, TYPE_SPINNER, EDITTYPE_UNIVERSE, nextID(), nextID(), SPIN_AUTOSCALE,
	PB_END,
	pb_animated, _T("animated"), TYPE_BOOL, 0, 0,
		p_default, FALSE,
		p_ui, TYPE_SINGLECHEKBOX, nextID(),
	PB_END,
PB_END
);

//************************************************************
// VRayMeshGeom implementation
//************************************************************

VRayMeshGeom::VRayMeshGeom() {
	static int pblockDesc_inited=false;
	if (!pblockDesc_inited) {
		initPBlockDesc(param_blk);
		pblockDesc_inited=true;
	}
	pblock2=NULL;
	classDesc.MakeAutoParamBlocks(this);
	vassert(pblock2);
	suspendSnap=FALSE;
}

VRayMeshGeom::~VRayMeshGeom() {
}

void VRayMeshGeom::InvalidateUI() {
	param_blk.InvalidateUI(pblock2->LastNotifyParamID());
}

static Pb2TemplateGenerator templateGenerator;

void VRayMeshGeom::BeginEditParams(IObjParam *ip, ULONG flags, Animatable *prev) {
	// Automatic UI generation
	DLGTEMPLATE* tmp=templateGenerator.GenerateTemplate(pblock2, STR_DLGTITLE, 108);
	pmap=CreateCPParamMap2(pblock2, ip, hInstance, tmp, STR_DLGTITLE, 0);
	templateGenerator.ReleaseDlgTemplate(tmp);
}

void VRayMeshGeom::EndEditParams(IObjParam *ip, ULONG flags, Animatable *next) {
	DestroyCPParamMap2(pmap);
}

RefTargetHandle VRayMeshGeom::Clone(RemapDir& remap) {
	VRayMeshGeom* newob=new VRayMeshGeom();	
	BaseClone(this, newob, remap);
	newob->ReplaceReference(0, remap.CloneRef(pblock2));
	return newob;
}

Animatable* VRayMeshGeom::SubAnim(int i) {
	switch (i) {
		case 0: return pblock2;
		default: return NULL;
	}
}

TSTR VRayMeshGeom::SubAnimName(int i) {
	switch (i) {
		case 0: return STR_DLGTITLE;
		default: return _T("");
	}
}

RefTargetHandle VRayMeshGeom::GetReference(int i) {
	switch (i) {
		case REFNO_PBLOCK: return pblock2;
		default: return NULL;
	}
}

void VRayMeshGeom::SetReference(int i, RefTargetHandle rtarg) {
	switch (i) {
		case REFNO_PBLOCK: pblock2 = (IParamBlock2*)rtarg; break;
	}
}

RefResult VRayMeshGeom::NotifyRefChanged(NOTIFY_REF_CHANGED_ARGS) {
	switch (message) {
		case REFMSG_CHANGE:
			if (hTarget==pblock2) {
				// if (pblock2->LastNotifyParamID()==VRayMeshGeom_fileName) readPreview();
				param_blk.InvalidateUI();
			}
			break;
	}
	return REF_SUCCEED;
}

Interval VRayMeshGeom::ObjectValidity(TimeValue t) {
	Interval valid=FOREVER;
	float f;
	pblock2->GetValue(pb_radius, t, f, valid);
	return valid;
}

//**************************************************************************
int VRayMeshGeomCreateCallBack::proc(ViewExp *vpt, int msg, int point, int flags, IPoint2 m, Matrix3& mat) {
	if (!sphere) return CREATE_ABORT;

	Point3 np=vpt->SnapPoint(m,m,NULL,SNAP_IN_PLANE);

	switch (msg) {
		case MOUSE_POINT:
			switch (point) {
				case 0:
					sphere->suspendSnap=TRUE;				
					sp0=m;
					p0=vpt->SnapPoint(m, m, NULL, SNAP_IN_3D);
					mat.SetTrans(p0);
					sphere->pblock2->SetValue(pb_boxSize, 0, 0.0f);
					return CREATE_CONTINUE;
				case 1:
					float r=Length(np-mat.GetTrans());
					if (r<1e-3f) return CREATE_ABORT;

					sphere->pblock2->SetValue(pb_boxSize, 0, r);
					return CREATE_STOP;
			}
			return CREATE_CONTINUE;

		case MOUSE_MOVE:
			if (point==1) {
				float r=Length(np-mat.GetTrans());
				sphere->pblock2->SetValue(pb_boxSize, 0, r);
			}
			return CREATE_CONTINUE;

		case MOUSE_ABORT:
			return CREATE_ABORT;
	}

	return CREATE_CONTINUE;
}

static VRayMeshGeomCreateCallBack createCB;

CreateMouseCallBack* VRayMeshGeom::GetCreateMouseCallBack() {
	createCB.SetObj(this);
	return &createCB;
}

void VRayMeshGeom::SetExtendedDisplay(int flags) {
	extDispFlags = flags;
}

void VRayMeshGeom::GetLocalBoundBox(TimeValue t, INode* inode, ViewExp* vpt, Box3& box) {
	box.Init();
	float radius=pblock2->GetFloat(pb_boxSize, t);
	box+=Point3(-radius, -radius, -radius);
	box+=Point3(radius, radius, radius);
}

void VRayMeshGeom::GetWorldBoundBox(TimeValue t, INode* inode, ViewExp* vpt, Box3& box) {
	if (!inode) return;
	Box3 localBox;
	GetLocalBoundBox(t, inode, vpt, localBox);
	box=localBox*(inode->GetObjectTM(t));
}

void VRayMeshGeom::GetDeformBBox(TimeValue t, Box3 &b, Matrix3 *tm, BOOL useSel) {
	if (!tm) GetLocalBoundBox(t, NULL, NULL, b);
	else {
		Box3 bbox;
		GetLocalBoundBox(t, NULL, NULL, bbox);
		b.Init();
		for (int i=0; i<8; i++) b+=(*tm)*bbox[i];
	}
}

int VRayMeshGeom::HitTest(TimeValue t, INode *node, int type, int crossing, int flags, IPoint2 *p, ViewExp *vpt) {
	static HitRegion hitRegion;
	DWORD	savedLimits;

	GraphicsWindow *gw=vpt->getGW();	
	Material *mtl=gw->getMaterial();
	MakeHitRegion(hitRegion, type, crossing, 4, p);

	gw->setRndLimits(((savedLimits = gw->getRndLimits())|GW_PICK)&~GW_ILLUM);
	gw->setHitRegion(&hitRegion);
	gw->clearHitCode();

	draw(t, node, vpt);

	gw->setRndLimits(savedLimits);
	
	if((hitRegion.type != POINT_RGN) && !hitRegion.crossing) return TRUE;
	return gw->checkHitCode();
}

void VRayMeshGeom::Snap(TimeValue t, INode* inode, SnapInfo *snap, IPoint2 *p, ViewExp *vpt) {
	if (suspendSnap) return;
}

inline void drawLine(GraphicsWindow *gw, Point3 &p0, Point3 &p1) {
	Point3 p[3]={ p0, p1 };
	gw->segment(p, TRUE);
}

inline void drawTriangle(GraphicsWindow *gw, Point3 &p0, Point3 &p1, Point3 &p2) {
	Point3 n=Normalize((p1-p0)^(p2-p0));
	Point3 p[4]={ p0, p1, p2 };
	Point3 uvw[4]={ Point3(0,0,0), Point3(0,0,0), Point3(0,0,0) };
	Point3 nrm[4]={ n, n, n };
	gw->triangleN(p, nrm, uvw);
}

inline void drawBBox(GraphicsWindow *gw, Box3 &b) {
	gw->setTransform(Matrix3(1));
	Point3 p[8];
	for (int i=0; i<8; i++) p[i]=b[i];
	gw->startSegments();

	drawLine(gw, p[0], p[1]);
	drawLine(gw, p[0], p[2]);
	drawLine(gw, p[0], p[4]);

	/*
	drawLine(gw, p[7], p[6]);
	drawLine(gw, p[7], p[5]);
	drawLine(gw, p[7], p[3]);
	*/
	gw->endSegments();
}

const float size=40.0f;

void VRayMeshGeom::draw(TimeValue t, INode *node, ViewExp *vpt) {
	GraphicsWindow *gw=vpt->getGW();

	Matrix3 tm=node->GetObjectTM(t);
	/*
	tm.NoScale();
	float scaleFactor=vpt->NonScalingObjectSize()*vpt->GetVPWorldWidth(tm.GetTrans())/(float)360.0;
	tm.Scale(Point3(scaleFactor,scaleFactor,scaleFactor));
	*/
	gw->setTransform(tm);

	Color color=Color(node->GetWireColor());
	if (node->IsFrozen()) color=GetUIColor(COLOR_FREEZE);
	else if (node->Selected()) color=GetUIColor(COLOR_SELECTION);
	gw->setColor(LINE_COLOR, color);

	/*
	Point3 p[6];
	p[0]=Point3(-iconSize, -iconSize, 0.0f);
	p[1]=Point3(iconSize, -iconSize, 0.0f);
	p[2]=Point3(iconSize, iconSize, 0.0f);
	p[3]=Point3(-iconSize, iconSize, 0.0f);
	p[4]=Point3(0.0f, 0.0f, 0.0f);
	p[5]=Point3(0.0f, 0.0f, iconSize);

	gw->setColor(FILL_COLOR, color);
	gw->startTriangles();

	drawTriangle(gw, p[0], p[1], p[2]);
	drawTriangle(gw, p[2], p[3], p[0]);

	gw->endTriangles();

	gw->setColor(LINE_COLOR, color);
	gw->startSegments();

	drawLine(gw, p[0], p[1]);
	drawLine(gw, p[1], p[2]);
	drawLine(gw, p[2], p[3]);
	drawLine(gw, p[3], p[0]);
	drawLine(gw, p[0], p[2]);
	drawLine(gw, p[1], p[3]);
	drawLine(gw, p[4], p[5]);
	drawLine(gw, p[5], p[5]*0.8f+p[0]*0.2f);
	drawLine(gw, p[5], p[5]*0.8f+p[1]*0.2f);
	drawLine(gw, p[5], p[5]*0.8f+p[2]*0.2f);
	drawLine(gw, p[5], p[5]*0.8f+p[3]*0.2f);

	gw->endSegments();
	*/

	float radius=pblock2->GetFloat(pb_boxSize, t);
	int nsegs=30;
	float u0=radius, v0=0.0f;
	Point3 pt[3];

	gw->startSegments();

	for (int i=0; i<nsegs; i++) {
		float a=2.0f*pi*float(i+1)/float(nsegs);

		float u1=radius*cosf(a);
		float v1=radius*sinf(a);

		pt[0]=Point3(u0, v0, 0.0f);
		pt[1]=Point3(u1, v1, 0.0f);
		gw->segment(pt, true);

		pt[0]=Point3(0.0f, u0, v0);
		pt[1]=Point3(0.0f, u1, v1);
		gw->segment(pt, true);

		pt[0]=Point3(u0, 0.0f, v0);
		pt[1]=Point3(u1, 0.0f, v1);
		gw->segment(pt, true);

		u0=u1;
		v0=v1;
	}

	gw->endSegments();

	/*
	if (hitTest && gw->checkHitCode()) {
		*res=TRUE;
		vpt->CtrlLogHit(inode, gw->getHitDistance(), 0, 0);
		if (abortOnHit) break;
		gw->clearHitCode();
	}
	*/

	tm.NoScale();
	float scaleFactor=vpt->NonScalingObjectSize()*vpt->GetVPWorldWidth(tm.GetTrans())/(float)360.0;
	tm.Scale(Point3(scaleFactor,scaleFactor,scaleFactor));
	gw->setTransform(tm);

	IPoint3 ipt;
	Point3 org(0,0,0);
	gw->wTransPoint(&org, &ipt);

	SIZE sp;
	gw->getTextExtents(iconText, &sp);

	ipt.x-=sp.cx/2;
	ipt.y-=sp.cy/2;

	gw->setColor(TEXT_COLOR, 0.0f, 0.0f, 0.0f);
	gw->wText(&ipt, iconText);

	ipt.x--;
	ipt.y--;
	gw->setColor(TEXT_COLOR, 1.0f, 1.0f, 1.0f);
	gw->wText(&ipt, iconText);
}

int VRayMeshGeom::Display(TimeValue t, INode* node, ViewExp *vpt, int flags) {
	draw(t, node, vpt);
	return 0;
}

ObjectState VRayMeshGeom::Eval(TimeValue time) {
	return ObjectState(this);
}

void* VRayMeshGeom::GetInterface(ULONG id) {
	if (id==I_VRAYGEOMETRY) return (VR::VRenderObject*) this;
	return GeomObject::GetInterface(id);
}

void VRayMeshGeom::ReleaseInterface(ULONG id, void *ip) {
	if (id==I_VRAYGEOMETRY) return;
	GeomObject::ReleaseInterface(id, ip);
}

Mesh* VRayMeshGeom::GetRenderMesh(TimeValue t, INode *inode, View& view, BOOL& needDelete) {
	needDelete=false;
	return &mesh;
}

//**************************************************************
// VRenderObject
int VRayMeshGeom::init(const ObjectState &os, INode *node, VR::VRayCore *vray) {
	VRenderObject::init(os, node, vray);
	return true;
}

// Generic plugin enumeration callback; used to implement the renderBegin/renderEnd/frameBegin/frameEnd
// callbacks for V-Ray shaders.
struct VRayCallback: EnumPluginCallback {
	VRayCallback(VR::VRayCore *vrayCore) { vray=static_cast<VR::VRayRenderer*>(vrayCore); }

	int process(Plugin* plugin) {
		VR::VRayPlugin *vrayPlugin=static_cast<VR::VRayPlugin*>(GET_INTERFACE(plugin, EXT_VRAY_PLUGIN));
		if (vrayPlugin) {
			processVRayPlugin(vrayPlugin);
		}
		return true;
	}

	virtual void processVRayPlugin(VR::VRayPlugin *vrayPlugin)=0;

protected:
	VR::VRayRenderer *vray;
};

// A plugin enumeration callback that calls renderBegin() on the V-Ray shaders.
struct RenderBeginCB: VRayCallback {
	RenderBeginCB(VR::VRayCore *vrayCore): VRayCallback(vrayCore) {}
	void processVRayPlugin(VR::VRayPlugin *p) { p->renderBegin(vray); }
};

// A plugin enumeration callback that calls renderEnd() on the V-Ray shaders.
struct RenderEndCB: VRayCallback {
	RenderEndCB(VR::VRayCore *vrayCore): VRayCallback(vrayCore) {}
	void processVRayPlugin(VR::VRayPlugin *p) { p->renderEnd(vray); }
};

// A plugin enumeration callback that calls frameBegin() on the V-Ray shaders.
struct FrameBeginCB: VRayCallback {
	FrameBeginCB(VR::VRayCore *vrayCore): VRayCallback(vrayCore) {}
	void processVRayPlugin(VR::VRayPlugin *p) { p->frameBegin(vray); }
};

// A plugin enumeration callback that calls frameEnd() on the V-Ray shaders.
struct FrameEndCB: VRayCallback {
	FrameEndCB(VR::VRayCore *vrayCore): VRayCallback(vrayCore) {}
	void processVRayPlugin(VR::VRayPlugin *p) { p->frameEnd(vray); }
};

void VRayMeshGeom::renderBegin(TimeValue t, VR::VRayCore *vray) {
	VRenderObject::renderBegin(t, vray);
	if (0==numRenderPlugins && plugman) {
		// Call renderBegin() on the V-Ray shaders
		RenderBeginCB cb(vray);
		plugman->enumPlugins(&cb);
	}
	numRenderPlugins++;
}

void VRayMeshGeom::renderEnd(VR::VRayCore *vray) {
	VRenderObject::renderEnd(vray);
	numRenderPlugins--;
	if (0==numRenderPlugins && plugman) {
		// Call renderEnd() on the V-Ray shaders
		RenderEndCB cb(vray);
		plugman->enumPlugins(&cb);
		clearShaders();
	}
}

void VRayMeshGeom::frameBegin(TimeValue t, VR::VRayCore *vray) {
	VRenderObject::frameBegin(t, vray);

	int numSpheres=pblock2->GetInt(pb_numSpheres);
	float radius=pblock2->GetFloat(pb_radius);
	int numSegments=pblock2->GetInt(pb_numSegments);
	float bbox=pblock2->GetFloat(pb_boxSize);

	float time=0.0f;
	if (pblock2->GetInt(pb_animated)) time=(float) t/(float) GetTicksPerFrame();

	meshGen.init(numSpheres, radius, numSegments, bbox*2.0f, time, isMoving, vray->getSequenceData().rayserver, vray->getSequenceData().threadManager);

	if (0==numFramePlugins && plugman) {
		// Call frameBegin() on the V-Ray shaders
		FrameBeginCB cb(vray);
		plugman->enumPlugins(&cb);
	}
	numFramePlugins++;
}

void VRayMeshGeom::frameEnd(VR::VRayCore *vray) {
	VRenderObject::frameEnd(vray);
	numFramePlugins--;
	if (0==numFramePlugins && plugman) {
		// Call frameEnd() on the V-Ray shaders
		FrameEndCB cb(vray);
		plugman->enumPlugins(&cb);
	}
}

// Get the path to the V-Ray plugins; use the V-Ray RT environment variable for this.
const tchar* getVRayPluginPath() {
	char pluginsDirVar[512];
	vutils_strcpy(pluginsDirVar, VRAYRT_PLUGINS);
	vutils_strcat(pluginsDirVar, "_");
	vutils_strcat(pluginsDirVar, PROCESSOR_ARCHITECTURE);

	const char *s = getenv(pluginsDirVar);
	if (s == NULL) {
		tchar str[512];
		sprintf(str, "Could not read V-Ray environment variable \"%s\". Please re-install.\n", pluginsDirVar);
		VUtils::debug(str);
		return ".";
	}
	return s;
}

// Unload the V-Ray shader DLLs and destroy the plugin manager. This is done when
// 3ds Max shuts down.
void VRayMeshGeom::maxShutdown(void *param, NotifyInfo *info) {
	if (!info || info->intcode!=NOTIFY_SYSTEM_SHUTDOWN2) return;

	if (plugman) {
		plugman->unloadAll();
		deleteDefaultPluginManager(plugman);
		plugman=NULL;
	}
}

// Create a new plugin manager if one does not exist. For the moment, we directly create
// a plugin manager; however in future V-Ray builds, V-Ray for 3ds Max will probably provide a plugin
// manager similar to V-Ray for Maya and V-Ray Standalone, so this code may need to change
// in the future.
void VRayMeshGeom::createPluginManager(VR::ProgressCallback *prog) {
	if (plugman)
		return;

	plugman=newDefaultPluginManager();
	const tchar *vrayPluginPath=getVRayPluginPath();
	plugman->loadLibraryFromPathCollection(vrayPluginPath, "/vray_*.dll", NULL, prog);

	if (!shutdownCallbackRegistered) {
		shutdownCallbackRegistered=true;
		RegisterNotification(maxShutdown, NULL, NOTIFY_SYSTEM_SHUTDOWN2);
	}
}

// Create a new V-Ray plugin; if necessary, first create a default plugin manager.
VR::VRayPlugin* VRayMeshGeom::newPlugin(const tchar *type, VR::ProgressCallback *prog) {
	if (!plugman) {
		createPluginManager(prog);
		if (!plugman)
			return NULL;
	}

	Plugin *plugin=plugman->newPlugin(type, NULL);
	return static_cast<VR::VRayPlugin*>(GET_INTERFACE(plugin, EXT_VRAY_PLUGIN));
}

// Clear all the V-Ray shaders.
void VRayMeshGeom::clearShaders(void) {
	if (!plugman)
		return;

	factory.clear();
	plugman->deleteAll();
}

// Create a material for the instance with the given render ID.
VR::VRayPlugin* VRayMeshGeom::createSomeMaterial(int renderID, VR::ProgressCallback *prog) {
	// Here we create a Multi material with a different sub material for each sphere
	// When there are many spheres this can be very slow and cosume huge amounts of memory
	// so limit the sub materials to some small number.
	int numSubMtls=VR::Min(pblock2->GetInt(pb_numSpheres),10);

	VR::VRayPlugin *multiMtl=newPlugin("MtlMulti", prog);

	VR::DefPluginListParam *subMtls=factory.saveInFactory(new VR::DefPluginListParam("mtls_list"));
	multiMtl->setParameter(subMtls);

	VR::DefIntListParam *subIDs=factory.saveInFactory(new VR::DefIntListParam("ids_list"));
	multiMtl->setParameter(subIDs);

	subMtls->setCount(numSubMtls);
	subIDs->setCount(numSubMtls);

	VR::Random rnd(renderID);
	for (int i=0; i<numSubMtls; i++) {
		VR::VRayPlugin *uvwgen=newPlugin("UVWGenChannel", prog);
		if (uvwgen) {
			uvwgen->setParameter(factory.saveInFactory(new VR::DefIntParam("uvw_channel", 1))); // We generate the first mapping channel
			uvwgen->setParameter(factory.saveInFactory(new VR::DefTransformParam("tex_transform", VR::TraceTransform(1)*4.0f)));
		}

		VR::VRayPlugin *checker=newPlugin("TexChecker", prog);
		if (checker) {
			VR::Color col(rnd.rnd(), rnd.rnd(), rnd.rnd());
			col=col*0.5f+VR::Color(0.5f, 0.5f, 0.5f);
			checker->setParameter(factory.saveInFactory(new VR::DefColorParam("white_color", col)));
			checker->setParameter(factory.saveInFactory(new VR::DefColorParam("black_color", col*0.25f)));
			checker->setParameter(factory.saveInFactory(new VR::DefPluginParam("uvwgen", uvwgen)));
		}

		VR::VRayPlugin *brdf=newPlugin("BRDFDiffuse", prog);
		if (brdf) {
			brdf->setParameter(factory.saveInFactory(new VR::DefPluginParam("color_tex", checker)));
		}

		VR::VRayPlugin *mtl=newPlugin("MtlSingleBRDF", prog);
		if (mtl) {
			mtl->setParameter(factory.saveInFactory(new VR::DefPluginParam("brdf", brdf)));
		}

		subMtls->setPlugin(i, mtl);
		(*subIDs)[i]=i;
	}

	return multiMtl;
}

// Create a new render instance.
VR::VRenderInstance* VRayMeshGeom::newRenderInstance(INode *node, VR::VRayCore *vray, int renderID) {
	// Determine if we need motion blur
	isMoving=pblock2->GetInt(pb_animated) && vray->getSequenceData().params.moblur.on;

	// Create some material for our instance
	VR::VRayPlugin *mtlPlugin=createSomeMaterial(renderID, vray->getSequenceData().progress);

	return new VRayMeshGeomInstance(this, node, vray, renderID, isMoving, mtlPlugin);
}

void VRayMeshGeom::deleteRenderInstance(VR::VRenderInstance *ri) {
	delete (VRayMeshGeomInstance*) ri;
}

//**************************************************************
// The mesh generator

MeshTest::MeshTest(void) {
	size=1.0f;
	subdivs=10;
	numSpheres=0;
	time=0.0f;
	voxels=NULL;
	voxelInfo=NULL;
}

MeshTest::~MeshTest(void) {
	delete[] voxels;
	voxels=NULL;
	delete[] voxelInfo;
	voxelInfo=NULL;
}

void MeshTest::init(int n, float f, int s, float bsize, float t, int moving, VR::RayServerInterface2 *_rayserver, VR::ThreadManager *_threadman) {
	size=f;
	subdivs=s;
	bboxSize=bsize;
	time=t;
	isMoving=moving;
	rayserver=_rayserver;
	threadman=_threadman;

	numSpheres=n;
	delete[] voxels;
	delete[] voxelInfo;
	voxels=new VR::MeshVoxel[numSpheres];
	voxelInfo=new VoxelInfo[numSpheres];
}

// From MeshInterface
int MeshTest::getNumVoxels(void) {
	return numSpheres;
}

VR::Box MeshTest::getVoxelBBox(int i) {
	float radius;
	VR::Vector center;

	if (i!=0)
		getRandomSphere(i, radius, center, time);
	else {
		radius=1.0f;
		center.makeZero();
	}

	VR::Vector d(radius, radius, radius);

	VR::Box b;
	b.init();
	b+=center-d;
	b+=center+d;

	if (isMoving && i!=0) {
		float radius1;
		VR::Vector center1;
		getRandomSphere(i, radius1, center1, time+1.0f);

		VR::Vector d1(radius1, radius1, radius1);

		b+=center1-d1;
		b+=center1+d1;
	}

	return b;
}

uint32 MeshTest::getVoxelFlags(int i) {
	if (i!=0)
		return MVF_GEOMETRY_VOXEL;
	else
		// make the 0th voxel hidden and render only the instances
		return MVF_GEOMETRY_VOXEL | MVF_HIDDEN_VOXEL;
}

VR::MeshVoxel* MeshTest::getVoxel(int i, uint64 *memUsage=NULL) {
	VoxelInfo &vi=voxelInfo[i];

	vi.loadedCount++; // First increment the usage count, so that no-one deletes the voxel

	uint64 memUsed=0;
	if (!vi.isLoaded) {
		csect.enter();
		if (!vi.isLoaded) {
			buildVoxel(i, voxels[i]);
			memUsed=voxels[i].getMemUsage();
			vi.isLoaded=true;
		}
		csect.leave();
	}

	if (memUsage) *memUsage=memUsed;
	return &voxels[i];
}

void MeshTest::releaseVoxel(VR::MeshVoxel *voxel, uint64 *memUsage=NULL) {
	int i=voxel->index;
	VoxelInfo &vi=voxelInfo[i];

	uint64 memUsed=0;
	if (vi.loadedCount.decrement()==0) {
		csect.enter();
		if (vi.loadedCount.get()==0 && vi.isLoaded) {
			memUsed=voxels[i].getMemUsage();
			voxels[i].freeMem();
			vi.isLoaded=false;
		}
		csect.leave();
	}

	if (memUsage) *memUsage=memUsed;
}

void MeshTest::buildVoxel(int voxelIndex, VR::MeshVoxel &mesh) {
	// Create one voxel per sphere.
	// Since the only differences between the spheres are their position and size we can use
	// instancing:
	// voxel 0 will be our source voxel and it will contain all the necessary stuff - vertices,faces etc..
	// the sphere in it will be at coordinates (0,0,0) and with radius 1.0f
	// all other voxels will be instanced voxels with reference to voxel0 and TMS_CHANNEL for transforms
	// the transform itself will scale the spheres to the required radius and offset to the sphere possition
	// For now motion blur is not handled !
	if (voxelIndex==0) {
		// First prepare the channels
		int numChannels=7; // vertices, faces; normals, face normals; uvw vertices, uvw faces; face info.

		//if (isMoving) numChannels++; // Vertex velocity

		mesh.init();
		mesh.channels=new VR::MeshChannel[numChannels];
		mesh.numChannels=0;
		mesh.index=voxelIndex;
		mesh.setRayServer(rayserver);

		// The vertices
		int numVerts=VR::sqr(subdivs+1);

		VR::MeshChannel &geomChan=mesh.channels[mesh.numChannels++];
		geomChan.init(sizeof(VR::VertGeomData), numVerts, VERT_GEOM_CHANNEL, FACE_TOPO_CHANNEL, MF_VERT_CHANNEL);
		VR::VertGeomData *verts=(VR::VertGeomData*) geomChan.data;

		// The face topology
		int numFaces=VR::sqr(subdivs)*2;

		VR::MeshChannel &topoChan=mesh.channels[mesh.numChannels++];
		topoChan.init(sizeof(VR::FaceTopoData), numFaces, FACE_TOPO_CHANNEL, 0, MF_TOPO_CHANNEL);
		VR::FaceTopoData *faces=(VR::FaceTopoData*) topoChan.data;

		// The vertex normals
		VR::MeshChannel &vertNormalsChan=mesh.channels[mesh.numChannels++];
		vertNormalsChan.init(sizeof(VR::VertGeomData), numVerts, VERT_NORMAL_CHANNEL, VERT_NORMAL_TOPO_CHANNEL /* (1) */, MF_VERT_CHANNEL);
		VR::VertGeomData *normals=(VR::VertGeomData*) vertNormalsChan.data;

		// The face normals
		// In this case we use a separate normal faces channel, but this can be optimized if it is the same as the geometry faces
		// - in that case we can simply specify FACE_TOPO_CHANNEL in (1) above, instead of VERT_NORMAL_TOPO_CHANNEL
		VR::MeshChannel &faceNormalsChan=mesh.channels[mesh.numChannels++];
		faceNormalsChan.init(sizeof(VR::FaceTopoData), numFaces, VERT_NORMAL_TOPO_CHANNEL, 0, MF_TOPO_CHANNEL);
		VR::FaceTopoData *normalFaces=(VR::FaceTopoData*) faceNormalsChan.data;

		// The texture vertices (for mapping channel 1)
		VR::MeshChannel &uvwVertsChan=mesh.channels[mesh.numChannels++];
		uvwVertsChan.init(sizeof(VR::VertGeomData), numVerts, VERT_TEX_CHANNEL0+1, VERT_TEX_TOPO_CHANNEL0+1 /* (2) */, MF_VERT_CHANNEL);
		VR::VertGeomData *uvw=(VR::VertGeomData*) uvwVertsChan.data;

		// The texture faces (for mapping channel 1)
		// In this case we use a separate uvw faces channel, but this can be optimized if it is the same as the geometry faces
		// - in that case we can simply specify FACE_TOPO_CHANNEL in (2) above, instead of VERT_TEX_TOPO_CHANNEL0+1
		VR::MeshChannel &uvwFacesChan=mesh.channels[mesh.numChannels++];
		uvwFacesChan.init(sizeof(VR::FaceTopoData), numFaces, VERT_TEX_TOPO_CHANNEL0+1, 0, MF_TOPO_CHANNEL);
		VR::FaceTopoData *uvwFaces=(VR::FaceTopoData*) uvwFacesChan.data;

		// Face info channel
		VR::MeshChannel &faceInfoChan=mesh.channels[mesh.numChannels++];
		faceInfoChan.init(sizeof(VR::FaceInfoData), numFaces, FACE_INFO_CHANNEL, 0, MF_INFO_CHANNEL);
		VR::FaceInfoData *infoData=(VR::FaceInfoData*) faceInfoChan.data;

		// Get a random sphere.
		float radius;
		VR::Vector center;
		getRandomSphere(voxelIndex, radius, center, time);
		// quick and dirt fix
		// make the first sphere always at 0,0,0 and with radius 1.0f. The rest of the spheres will
		// instance this one and use TMS_CHANNEL to translate/scale the sphere
		center.makeZero();
		radius=1.0f;

		// Next build the actual data
		for (int i=0; i<=subdivs; i++) {
			for (int j=0; j<=subdivs; j++) {
				float u=float(i)/float(subdivs);
				float v=float(j)/float(subdivs);

				float fx=-1.0f+2.0f*u;
				float fy=v;

				VR::Vector dir=VR::getSphereCoords(fx*VR::pi()*0.5f, fy*VR::pi()*2.0f);

				int index=i*(subdivs+1)+j;
				verts[index]=center+dir*radius;
				normals[index]=dir;
				uvw[index].set(u, v, 0.0f);
			}
		}

		for (int i=0; i<subdivs; i++) {
			for (int j=0; j<subdivs; j++) {
				int fidx=i*subdivs+j;
				int vidx=i*(subdivs+1)+j;

				setFace(faces[fidx*2], vidx, vidx+subdivs+2, vidx+subdivs+1);
				setFace(faces[fidx*2+1], vidx, vidx+1, vidx+subdivs+2);

				setFace(normalFaces[fidx*2], vidx, vidx+subdivs+2, vidx+subdivs+1);
				setFace(normalFaces[fidx*2+1], vidx, vidx+1, vidx+subdivs+2);

				setFace(uvwFaces[fidx*2], vidx, vidx+subdivs+2, vidx+subdivs+1);
				setFace(uvwFaces[fidx*2+1], vidx, vidx+1, vidx+subdivs+2);
			}
		}

		// Set all the material IDs to be the same as the sphere index
		for (int j=0; j<numFaces; j++) {
			 infoData[j].mtlID=voxelIndex;
		}

		// Add velocity information only if you want all instances to have that motion
		// Instances will also have transform motion added on top of this
		//if (isMoving) {
		//	// Compute the velocity for one frame; in our case, assume that the entire voxel has the same velocity
		//	float radius1;
		//	VR::Vector center1;
		//	getRandomSphere(voxelIndex, radius1, center1, time+1.0f);
		//	VR::Vector velocity=center1-center;

		//	// Vertex velocity channel
		//	VR::MeshChannel &vertVelChan=mesh.channels[mesh.numChannels++];
		//	vertVelChan.init(sizeof(VR::VertGeomData), numVerts, VERT_VELOCITY_CHANNEL, FACE_TOPO_CHANNEL, MF_VERT_CHANNEL);
		//	VR::VertGeomData *velData=(VR::VertGeomData*) vertVelChan.data;

		//	for (int i=0; i<numVerts; i++) {
		//		velData[i]=velocity;
		//	}
		//}
	}
	else {
		// Instance voxels - they have only a source voxel index and a transform
		int numChannels=2; // source index

		mesh.init();
		mesh.channels=new VR::MeshChannel[numChannels];
		mesh.numChannels=0;
		mesh.index=voxelIndex;
		mesh.setRayServer(rayserver);

		VR::MeshChannel &srcChan=mesh.channels[mesh.numChannels++];
		srcChan.init(sizeof(int), 1, SOURCE_VOXEL_INDEX_CHANNEL, 0, MF_INFO_CHANNEL);
		reinterpret_cast<int*>(srcChan.data)[0]=0;

		// Here we provide 2 transforms - for current time and time+1
		VR::MeshChannel &tmChan=mesh.channels[mesh.numChannels++];
		tmChan.init(sizeof(VR::TraceTransform), 2, TMS_CHANNEL, 0, MF_INFO_CHANNEL);
		VR::TraceTransform *tms = reinterpret_cast<VR::TraceTransform*>(tmChan.data);

		float radius;
		VR::Vector center;
		getRandomSphere(voxelIndex, radius, center, time);

		tms[0].makeIdentity();
		tms[0]*=radius;
		tms[0].offs=VR::TracePoint(center);


		float radius1;
		VR::Vector center1;
		getRandomSphere(voxelIndex, radius1, center1, time+1.0f);

		tms[1].makeIdentity();
		tms[1]*=radius1;
		tms[1].offs=VR::TracePoint(center1);
	}
}

void MeshTest::setFace(VR::FaceTopoData &f, int v0, int v1, int v2) {
	f.v[0]=v0;
	f.v[1]=v1;
	f.v[2]=v2;
}

void MeshTest::getRandomSphere(int seed, float &radius, VR::Vector &center, float time) {
	VR::Random rnd(seed);
	radius=VR::interpolate(rnd.rnd(), size*0.5f, size);

	VR::Vector orbitNormal=VR::getSphereDir(rnd.rnd()*2.0f-1.0f, rnd.rnd());

	float orbitDistance=rnd.rnd()*bboxSize*0.5f;
	static const float mult=10.0f;
	float orbitSpeed=mult*rnd.rnd()*(10.0f*VR::pi()/180.0f); // Maximum speed 10 degrees per frame
	float orbitOffset=rnd.rnd()*2.0f*VR::pi();

	float orbitAngle=orbitOffset+time*orbitSpeed;

	// Compute the sphere position
	VR::Matrix nm;
	VR::makeNormalMatrix(orbitNormal, nm);

	float cs=cosf(orbitAngle);
	float sn=sinf(orbitAngle);

	VR::Vector orbitDir=(nm[0]*cs+nm[1]*sn);
	center=orbitDir*orbitDistance;
}
